from ..args import PathArgument, convert_path_arguments
from ..modules import module_path


class SageMakerProcessingConfig(object):
    def __init__(self, dependencies=None, inputs=None, outputs=None, env=None):
        self.inputs = convert_path_arguments(inputs)
        self.outputs = convert_path_arguments(outputs)
        self.dependencies = dependencies or {}
        self.dependencies = {
            k: module_path(v)
            for k, v in self.dependencies.items()
        }
        self.env = env or {}



from scipy.io import wavfile
import io

def input_fn_wav(request_body, request_content_type):
    if request_content_type in ['audio/wav', 'audio/wave']:
        fs, data = wavfile.read(io.BytesIO(request_body))
        return fs, data
    else:
        raise ValueError(
            "Unsupported content type: {}".format(request_content_type))



import sagemaker
import click
from botocore.exceptions import ClientError
from aws_sagemaker_remote.util.training import training_describe
from .iam import ensure_inference_role
from aws_sagemaker_remote.util.fields import get_field


def model_delete(name, client):
    client.delete_model(ModelName=name)


def model_describe(name, client, field=None):
    description = client.describe_model(ModelName=name)
    description = get_field(data=description, field=field)
    return description


def model_exists(name, client):
    try:
        model_describe(name=name, client=client)
        return True
    except ClientError:
        return False


def model_create(
    job, model_artifact,
    name, session:sagemaker.Session, inference_image, role, force,
    accelerator_type=None
):
    if (job and model_artifact) or (not (job or model_artifact)):
        raise click.UsageError('Specify one of job_name or model_artifact')
    if model_artifact and not name:
        raise click.UsageError(
            'name is required if job is not provided')
    iam = session.boto_session.client('iam')
    client = session.boto_session.client('sagemaker')
    role = ensure_inference_role(iam=iam, role_name=role)
    if job:
        client = session.boto_session.client('sagemaker')
        model_artifact = training_describe(
            job_name=job,
            field='ModelArtifacts.S3ModelArtifacts',
            client=client
        )
        if not name:
            name = job
        print("Creating model [{}] from job [{}] artifact [{}]".format(
            name, job, model_artifact))
    else:
        if not model_artifact.startswith('s3://'):
            if model_artifact.startswith('/'):
                model_artifact=model_artifact[1:]
            bucket=session.default_bucket()
            model_artifact='s3://{}/{}'.format(bucket, model_artifact)
        print("Creating model [{}] from artifact [{}]".format(
            name, model_artifact))

    if model_exists(name=name, client=client):
        if force:
            print("Deleting existing model")
            model_delete(name=name, client=client)
        else:
            raise click.UsageError('Specify force if overwriting model')
    model = sagemaker.Model(
        image_uri=inference_image,
        model_data=model_artifact,
        role=role,
        predictor_cls=None,
        env=None,
        name=name,
        # vpc_config=None,
        sagemaker_session=session,
        # enable_network_isolation=False,
        # model_kms_key=None
    )
    container_def = sagemaker.container_def(
        model.image_uri,
        model.model_data,
        model.env)

    # self._ensure_base_name_if_needed(container_def["Image"])
    # self._set_model_name_if_needed()

    enable_network_isolation = model.enable_network_isolation()

    # self._init_sagemaker_session_if_does_not_exist(instance_type)
    session.create_model(
        model.name,
        model.role,
        container_def,
        vpc_config=model.vpc_config,
        enable_network_isolation=enable_network_isolation,
        # tags=tags,
    )



import click
import os
import sys


def inference_local(model_dir, input, output, input_type, output_type):
    try:
        from sagemaker_inference.default_handler_service import DefaultHandlerService
    except ImportError:
        raise click.UsageError(
            "Install sagemaker-inference to use local inference")

    try:
        from mms.context import Context, RequestProcessor
    except ImportError:
        raise click.UsageError(
            "Install multi-model-server to use local inference")

    sys.path.insert(0, os.path.join(model_dir, 'code'))

    with open(input, 'rb') as f:
        data = f.read()
    handler = DefaultHandlerService()
    context = Context(
        model_name='local-model',
        model_dir=model_dir,
        manifest=None,
        batch_size=None,
        gpu=None,
        mms_version=None
    )
    context.request_processor = [RequestProcessor(
        request_header={
            'Content-Type': input_type,
            'Accept': output_type
        }
    )]
    data = [
        {'body': data}
    ]
    handler.initialize(context)
    result = handler.handle(data, context)
    #print("Result: {}".format(result))
    status = context.get_response_status(0)
    #print(status)
    if status[0] != 200 and status[0].value != 200:
        raise ValueError("Invokation error: {}\n{}".format(status, "\n".join(result)))
    assert len(result) > 0
    if output:
        os.makedirs(os.path.dirname(os.path.abspath(output)), exist_ok=True)
        if isinstance(result[0], str):
            mode = 'w'
        elif isinstance(result[0], bytes):
            mode='wb'
        else:
            raise ValueError("Unknown result type: {}".format(type(result[0])))
        with open(output, mode) as f:
            for chunk in result:
                f.write(chunk)
        print("Saved to {}".format(output))
        return None
    else:
        return result
    # return "data"



from .process import sagemaker_processing_run
from .args import sagemaker_processing_args
from argparse import ArgumentParser
import inspect
from ..commands import Command, run_command


class ProcessingCommand(Command):
    def __init__(
            self,
            script, main, help=None,
            **processing_args
    ):
        super(ProcessingCommand, self).__init__(
            help=help or 'SageMaker processing'
        )

        if not script:
            script = inspect.getfile(main)
        self.script = script
        self.main = main
        self.processing_args = processing_args
        self.config = None

    def configure(self, parser: ArgumentParser):
        self.config = sagemaker_processing_args(
            parser=parser,
            script=self.script,
            **self.processing_args)

    def run(self, args):
        sagemaker_processing_handle(
            args=args,
            config=self.config,
            main=self.main
        )


def sagemaker_processing_handle(
    args, config, main
):
    if args.sagemaker_run:
        # Remote processing
        sagemaker_processing_run(
            args=args,
            config=config
        )
    else:
        # Local processing
        main(args)


def sagemaker_processing_main(
    main, script=None, description=None,
    **processing_args
):
    r"""
    Entry point for processing.

    Example
    -------

    .. code-block:: python

        from aws_sagemaker_remote import sagemaker_training_main

        def main(args):
            # your code here
            pass

        if __name__ == '__main__':
            sagemaker_training_main(
                main=main,
                # ... additional configuration
            )

    Parameters
    ----------
    main : function
        Main function. Must accept a single argument ``args:argparse.Namespace``.
    script : str, optional
        Path to script file to execute. 
        Set to ``__file__`` for most use-cases.
        Empty or None defaults to file containing ``main``.
    description: str, optional
        Script description for argparse
    \**processing_args : dict, optional
        Keyword arguments to :meth:`aws_sagemaker_remote.processing.args.sagemaker_processing_args`
    """
    command = ProcessingCommand(
        script=script,
        main=main,
        **processing_args
    )
    run_command(command, description=description)



import argparse
from distutils.util import strtobool
import os
from .config import SageMakerProcessingConfig
from ..args import variable_to_argparse, bool_argument, sagemaker_profile_args, PROFILE


PROCESSING_ROLE = 'aws-sagemaker-remote-processing-role'
PROCESSING_IMAGE = '683880991063.dkr.ecr.us-east-1.amazonaws.com/columbo-compute:latest'
PROCESSING_INSTANCE = 'ml.t3.medium'
PROCESSING_JOB_NAME = 'processing-job'
PROCESSING_RUNTIME_SECONDS = 60*60  # 1 hour
PROCESSING_VOLUME_SIZE = 30  # GB

MODULE_MOUNT = '/opt/ml/processing/modules'
MODULE_MOUNT_HELP = """Mount point for modules.
If running on SageMaker, modules are mounted here and this directory is added to PYTHONPATH
(default: [{}])
"""
MODULE_HELP = """Directory of [{k}] module.
If running on SageMaker, modules are uploaded and placed on PYTHONPATH.
(default: [{v}])
"""

INPUT_MOUNT = '/opt/ml/processing/input'
INPUT_MOUNT_HELP = """Mount point for inputs.
If running on SageMaker, inputs are mounted here.
If running locally, S3 inputs are downloaded here. 
No effect on local inputs when running locally.
(default: [{}])
"""
INPUT_HELP = """Input [{k}].
Local path or path on S3.
If running locally, local paths are used directly.
If running locally, S3 paths are downloaded to [`--sagemaker-input-mount`/{k}].
If running on SageMaker, local paths are uploaded to S3 then S3 data is downloaded to [`--sagemaker-input-mount`/{k}].
If running on SageMaker, S3 paths are downloaded to [`--sagemaker-input-mount`/{k}].
(default: [{v}])
"""
INPUT_MODE_HELP = """Input [{k}] mode. File or Pipe.
(default: [{v}])
"""

OUTPUT_MOUNT = '/opt/ml/processing/output'
OUTPUT_MOUNT_HELP = """Mount point for outputs.
If running on SageMaker, outputs written here are uploaded to S3.
If running locally, S3 outputs written here are uploaded to S3. 
No effect on local outputs when running locally.
(default: [{}])
"""
OUTPUT_HELP = """Output [{k}] local path.
If running locally, set to a local path.
(default: [{v}])
"""
OUTPUT_S3_HELP = """Output [{k}] S3 URI.
Upload results to this URI. Empty string automatically generates a URI.
(default: [{v}])
"""
OUTPUT_MODE_HELP = """Output [{k}] mode.
Set to Continuous or EndOfJob.
(default: [{v}])
"""


def sagemaker_processing_input_args(parser: argparse.ArgumentParser, inputs=None,
                                    input_mount=INPUT_MOUNT):
    if inputs is None:
        inputs = {}
    if len(inputs) > 0:
        group = parser.add_argument_group(
            title='Inputs',
            description='Input options'
        )
        group.add_argument(
            '--sagemaker-input-mount',
            default=input_mount,
            help=INPUT_MOUNT_HELP.format(input_mount))
        for k, v in inputs.items():
            flag = variable_to_argparse(k)
            group.add_argument(
                flag,
                default=v.local,
                help=INPUT_HELP.format(k=k, v=v.local))
            mode = v.mode or 'File'
            group.add_argument(
                variable_to_argparse("{}_mode".format(k)),
                default=mode,
                help=INPUT_MODE_HELP.format(k=k, v=mode))


def sagemaker_processing_output_args(parser: argparse.ArgumentParser, outputs=None,
                                     output_mount=OUTPUT_MOUNT):
    if outputs is None:
        outputs = {}
    if len(outputs) > 0:
        group = parser.add_argument_group(
            title='Output',
            description='Output options'
        )
        group.add_argument(
            '--sagemaker-output-mount',
            default=output_mount,
            help=OUTPUT_MOUNT_HELP.format(output_mount))
        for k, v in outputs.items():
            group.add_argument(
                variable_to_argparse(k),
                default=v.local,
                help=OUTPUT_HELP.format(k=k, v=v.local))
            group.add_argument(
                variable_to_argparse("{}_s3".format(k)),
                default=v.remote,
                help=OUTPUT_S3_HELP.format(k=k, v=v.remote))
            mode = v.mode or 'EndOfJob'
            group.add_argument(
                variable_to_argparse("{}_mode".format(k)),
                default=mode,
                help=OUTPUT_MODE_HELP.format(k=k, v=mode))


def sagemaker_processing_module_args(parser: argparse.ArgumentParser, dependencies=None,
                                     module_mount=MODULE_MOUNT):
    if dependencies is None:
        dependencies = {}
    group = parser.add_argument_group(
        title='Modules', description='Module options'
    )
    group.add_argument(
        '--sagemaker-module-mount',
        default=module_mount,
        help=MODULE_MOUNT_HELP.format(module_mount))
    for k, v in dependencies.items():
        flag = variable_to_argparse(k)
        group.add_argument(
            flag,
            default=v,
            help=MODULE_HELP.format(k=k, v=v))


def sagemaker_processing_args(
    parser: argparse.ArgumentParser,
    script,
    run=False,
    wait=True,
    profile=PROFILE,
    role=PROCESSING_ROLE,
    image=PROCESSING_IMAGE,
    instance=PROCESSING_INSTANCE,
    inputs=None,
    outputs=None,
    dependencies=None,
    input_mount=INPUT_MOUNT,
    output_mount=OUTPUT_MOUNT,
    module_mount=MODULE_MOUNT,
    base_job_name=PROCESSING_JOB_NAME,
    job_name='',
    runtime_seconds=PROCESSING_RUNTIME_SECONDS,
    volume_size=PROCESSING_VOLUME_SIZE,
    python='python3',
    requirements=None,
    configuration_script=None,
    configuration_command=None,
    additional_arguments=None,
    argparse_callback=None,
    output_json=None,
    env=None
):
    r"""
    Configure ``argparse.ArgumentParser`` for processing scripts.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        Parser to configure
    script : str
        Path to script file to execute.
        Set default for ``--sagemaker-script``
    run : bool, optional
        Run on SageMaker. 
        Set default for ``--sagemaker-run``.
    wait : bool, optional
        Wait for SageMaker processing to complete. 
        Set default for ``--sagemaker-wait``.
    profile : str, optional
        AWS profile to use for session. 
        Set default for ``--sagemaker-profile``.
    role : str, optional
        AWS IAM role name to use for processing. Will be created if it does not exist. 
        Set default for ``--sagemaker-role``.
    image : str, optional
        URI of ECR Docker image to use for processing. 
        Set default for ``--sagemaker-image``.
    instance : str, optional
        Type of instance to use for processing (e.g., ``ml.t3.medium``). 
        Set default for ``--sagemaker-instance``.
    inputs : dict(str,str), optional
        Dictionary of input argument keys to strings or :class:`aws_sagemaker_remote.args.PathArgument`.
        Strings are converted to ``PathArgument`` with ``local`` set to your string.
        This should be sufficient for most use cases.
        For eack key and value, create an argument ``--key`` that defaults to value.

        * Running locally, input arguments are unmodified.
        * Running remotely, inputs are set to appropriate SageMaker mount points. Local inputs are uploaded automatically.

        For example:

        .. code-block:: python

           import OPTIONAL, PathArgument from aws_sagemaker_remote.args
           inputs = {
               "my_input_1": "path/to/data1", # implicit
               "my_input_2": PathArgument(local="path/to/data2"), # explicit
               "my_optional_input": OPTIONAL
           }

        Your script will now have arguments ``--my-input-1``, ``--my-input-2``, and ``--my-optional-input``.

    outputs : dict(str, str)
        Dictionary of output arguments keys to strings or :class:`aws_sagemaker_remote.args.PathArgument`.
        Strings are converted to ``PathArgument`` with ``local`` set to your string.
        This should be sufficient for most use cases.
        For eack key and value, create an argument ``--key`` that defaults to value.

        For eack key:

            * Create an argument ``--key`` that defaults to ``value.local``. This controls an output path.
            * Create an argument ``--key-s3`` that defaults to ``value.remote``. This controls where output is stored on S3.
              * Set to ``default`` to automatically create an output path based on the job name
              * Set to an S3 URL to store output at a specific location on S3
    dependencies : dict(str, str)
        Dictionary of modules.
        For eack key and value, create an argument ``--module-key`` that defaults to value. 
        This controls the path of a dependency of your code.
        The files at the given path will be uploaded to S3, downloaded to SageMaker, and put on PYTHONPATH.    
    input_mount : str, optional
        Local path on SageMaker container where inputs are downloaded.
        Set default for ``--sagemaker-input-mount``.
    output_mount : str, optional
        Local path on SageMaker container where outputs are written before upload.
        Set default for ``--sagemaker-output-mount``.
    module_mount : str, optional
        Local path on SageMaker container where source code is downloaded. Mount point is put on PYTHONPATH.
        Set default for ``--sagemaker-module-mount``.
    base_job_name : str, optional
        Job name will be generated from ``base_job_name`` and a timestamp if ``job_name`` is not provided.
        Set default for ``--sagemaker-base-job-name``.
    job_name : str, optional
        Job name is used for tracking and organization. 
        Generated from ``base_job_name`` if not provided.
        Use ``base_job_name`` and leave ``job_name`` blank for most use-cases.
        Set default for ``--sagemaker-job-name``.
    runtime_seconds: int, optional
        Maximum in seconds before killing job.
        Set default for ``--sagemaker-runtime-seconds``.
    volume_size: int, optional
        Volume size in GB.
        Set default for ``--sagemaker-volume-size``.
    python: str, optional
        Pyton executable on container (default: ``python3``).
        Set default for ``--sagemaker-python``.
    requirements: str, optional
        Set path to requirements file to upload and install with ``pip install -r``.
        Set default for ``--sagemaker-requirements``.
    configuration_script: str, optional
        Set path to bash script to upload and source.
        Set default for ``--sagemaker-configuration-script``.
    configuration_command: str, optional
        Set command to be run to configure container,
        e.g. ``pip install mypackage && export MYVAR=MYVALUE``.
        Set default for ``--sagemaker-configuration-command``.
    additional_arguments: list, optional
        List of tuple of positional args and keyword args for ``argparse.ArgumentParser.add_argument``.
        Use to add additional arguments to the script.
    argparse_callback: function, optional
        Function accepting one argument ``parser:argparse.ArgumentParser`` that adds additional arguments.
        Use to add additional arguments to the script.
    output_json: str, optional
        Write SageMaker training details to this path.
        Set default for ``--sagemaker-output-json``
    """
    config = SageMakerProcessingConfig(
        dependencies=dependencies,
        inputs=inputs,
        outputs=outputs,
        env=env
    )

    if additional_arguments is None:
        additional_arguments = []
    group = parser.add_argument_group(
        title='SageMaker', description='SageMaker options'
    )
    sagemaker_profile_args(parser=group, profile=profile)
    bool_argument(group, '--sagemaker-run', default=run,
                  help="Run processing on SageMaker (yes/no default={})".format(run))
    bool_argument(group, '--sagemaker-wait', default=wait,
                  help="Wait for SageMaker processing to complete and tail logs (yes/no default={})".format(wait))
    group.add_argument('--sagemaker-script', default=script,
                       help='Python script to execute (default: [{}])'.format(script))
    group.add_argument('--sagemaker-python', default=python,
                       help='Python executable to use in container (default: [{}])'.format(python))
    group.add_argument('--sagemaker-job-name', default=job_name,
                       help='Job name for SageMaker processing. '
                       'If not provided, will be generated from base job name. '
                       'Leave blank for most use-cases. '
                       '(default: [{}])'.format(job_name))
    group.add_argument('--sagemaker-base-job-name', default=base_job_name,
                       help='Base job name for SageMaker processing .'
                       'Job name will be generated from the base name and a timestamp '
                       '(default: [{}])'.format(base_job_name))
    group.add_argument('--sagemaker-runtime-seconds', default=runtime_seconds, type=int,
                       help='SageMaker maximum runtime in seconds (default: [{}])'.format(runtime_seconds))
    group.add_argument('--sagemaker-role', default=role,
                       help='AWS role for SageMaker execution (default: [{}])'.format(role))
    group.add_argument('--sagemaker-requirements', default=requirements,
                       help='Requirements file to install on SageMaker (default: [{}])'.format(requirements))
    group.add_argument('--sagemaker-configuration-script', default=configuration_script,
                       help='Bash configuration script to source on SageMaker (default: [{}])'.format(configuration_script))
    group.add_argument('--sagemaker-configuration-command', default=configuration_command,
                       help='Bash command to run on SageMaker for configuration '
                       '(e.g., ``pip install aws_sagemaker_remote && export MYVAR=MYVALUE``) '
                       '(default: [{}])'.format(configuration_command))
    group.add_argument('--sagemaker-image', default=image,
                       help='AWS ECR image URI of Docker image to run SageMaker processing (default: [{}])'.format(image))
    group.add_argument('--sagemaker-instance', default=instance,
                       help='AWS SageMaker instance to run processing (default: [{}])'.format(instance))
    group.add_argument('--sagemaker-volume-size', default=volume_size, type=int,
                       help='AWS SageMaker volume size in GB (default: [{}])'.format(volume_size))
    group.add_argument('--sagemaker-output-json', default=output_json, type=str,
                       help='Write SageMaker training details to JSON file (default: [{}])'.format(output_json))
    sagemaker_processing_input_args(
        parser=parser,
        inputs=config.inputs,
        input_mount=input_mount
    )
    sagemaker_processing_output_args(
        parser=parser,
        outputs=config.outputs,
        output_mount=output_mount
    )
    sagemaker_processing_module_args(
        parser=parser,
        dependencies=config.dependencies,
        module_mount=module_mount
    )
    if argparse_callback:
        group = parser.add_argument_group(
            title='Processing', description='Processing options'
        )
        argparse_callback(group)
    # todo:
    # for args, kwargs in additional_arguments:
    #    parser.add_argument(*args, **kwargs)
    return config


def sagemaker_processing_parser_for_docs():
    parser = argparse.ArgumentParser()
    script = 'script.py'
    sagemaker_processing_args(
        parser=parser,
        script=script,
        inputs={
            'input': '/path/to/input'
        },
        outputs={
            'output': '/path/to/output'
        },
        dependencies={
            'my_module': '/path/to/my_module'
        }
    )
    return parser



import sagemaker
import os
import click
import re
from sagemaker.s3 import S3Uploader
import tarfile
from sagemaker.utils import _tmpdir
import boto3
from urllib.parse import urlparse


def upload(src, dst, gz, session: sagemaker.Session):
    if not os.path.exists(src):
        raise click.UsageError("Source must exist")
    if not dst.startswith('s3://'):
        if dst.startswith('/'):
            dst = dst[1:]
        bucket = session.default_bucket()
        dst = 's3://{}/{}'.format(bucket, dst)
    url = urlparse(dst)
    assert url.scheme == 's3'
    bucket = url.netloc
    key = url.path
    if key.startswith('/'):
        key = key[1:]
    if os.path.isfile(src):
        if gz:
            raise click.UsageError(
                "Option gz is only valid for source directories")
        s3 = session.boto_session.client('s3')
        s3.upload_file(src, bucket, key)
    elif os.path.isdir(src):
        if gz:
            if not re.match(".*\\.(tar\\.gz||tgz)$", dst, re.IGNORECASE):
                raise click.UsageError(
                    "Destination should end in .tar.gz or tgz")
            s3_dst = os.path.dirname(dst)
            file_name = os.path.basename(dst)
            with _tmpdir() as tmp:
                p = os.path.join(tmp, file_name)
                with tarfile.open(p, 'w:gz') as arc:
                    arc.add(name=src, arcname='.', recursive=True)
                s3 = session.boto_session.client('s3')
                s3.upload_file(p, bucket, key)
        else:
            S3Uploader.upload(
                local_path=src,
                desired_s3_uri=dst,
                sagemaker_session=session
            )
    else:
        raise click.UsageError("Source must be file or directory")



import sagemaker

import pprint
from aws_sagemaker_remote.util.fields import get_field

#logger = logging.getLogger()
#logger.addHandler(logging.StreamHandler()) # Writes to console
#logger.setLevel(logging.DEBUG)
#logging.getLogger('s3transfer').setLevel(logging.CRITICAL)
#logging.getLogger('urllib3').setLevel(logging.CRITICAL)

def processing_describe(job_name, client, field=None):
    description = processing_describe_get(job_name=job_name, client=client)
    description = get_field(description, field)
    return description


def processing_describe_get(client, job_name):
    description = client.describe_processing_job(
        ProcessingJobName=job_name
    )
    description['ProcessingInputs']={
        pi['InputName']: pi
        for pi in 
         description.get('ProcessingInputs', {})
    }
    description['ProcessingOutputConfig'] = description.get('ProcessingOutputConfig', {})
    description['ProcessingOutputConfig']['Outputs'] = {
        po['OutputName']: po
        for po in 
        description['ProcessingOutputConfig'].get('Outputs', {})
    }
    return description


if __name__ == '__main__':
    session = sagemaker.Session()
    client = session.boto_session.client('sagemaker')
    description = processing_describe(
        client=client, job_name='voxceleb-convert-vox2test-2020-10-07-09-23-16-900')
    print(description)


import sagemaker
import click
import botocore
from aws_sagemaker_remote.util.fields import get_field


def endpoint_config_describe(name, client, field=None):
    description = client.describe_endpoint_config(
        EndpointConfigName=name
    )
    description = get_field(data=description, field=field)
    return description


def endpoint_config_delete(name, client):
    response = client.delete_endpoint_config(
        EndpointConfigName=name
    )
    return response


def endpoint_config_exists(name, client):
    try:
        endpoint_config_describe(name=name, client=client)
        return True
    except botocore.exceptions.ClientError:
        return False


def endpoint_config_create(model, name, instance_type, force, session):
    if not model:
        raise click.UsageError("At least one model must be specified")
    if not name:
        if len(model) == 1:
            name = model[0]
        else:
            raise click.UsageError(
                "Must specifiy endpoint configuration name for multi-model endpoints")

    print("Creating endpoint configuration [{}] from models {}".format(
        name, model
    ))
    client = session.boto_session.client('sagemaker')
    if endpoint_config_exists(name=name, client=client):
        if force:
            print("Deleting existing endpoint config")
            endpoint_config_delete(name=name, client=client)
        else:
            raise click.UsageError(
                "Must specify `force` if overwriting existing endpoint config")

    variants = [
        {
            'VariantName': model_name,
            'ModelName': model_name,
            'InitialInstanceCount': 1,
            'InstanceType': instance_type,
            # 'InitialVariantWeight': ...,
            # 'AcceleratorType': 'ml.eia1.medium'|'ml.eia1.large'|'ml.eia1.xlarge'|'ml.eia2.medium'|'ml.eia2.large'|'ml.eia2.xlarge'
        }
        for model_name in model]
    response = client.create_endpoint_config(
        EndpointConfigName=name,
        ProductionVariants=variants
    )
    arn = response['EndpointConfigArn']
    print("Created endpoint configuration [{}]".format(arn))
    """
    DataCaptureConfig={
        'EnableCapture': True|False,
        'InitialSamplingPercentage': 123,
        'DestinationS3Uri': 'string',
        'KmsKeyId': 'string',
        'CaptureOptions': [
            {
                'CaptureMode': 'Input'|'Output'
            },
        ],
        'CaptureContentTypeHeader': {
            'CsvContentTypes': [
                'string',
            ],
            'JsonContentTypes': [
                'string',
            ]
        }
    },
    """
    # Tags=[
    #    {
    #        'Key': 'string',
    #        'Value': 'string'
    #    },
    # ],
    # KmsKeyId='string'
    # print(response)


if __name__ == '__main__':
    session = sagemaker.Session()
    client = session.boto_session.client('sagemaker')
    description = endpoint_config_describe_run(
        client=client, name='mnist-demo-2020-10-06-04-32-42-393')
    print(description)




from smexperiments import experiment


def ensure_experiment(client, experiment_name):
    try:
        return experiment.Experiment.load(
            experiment_name=experiment_name, sagemaker_boto_client=client)
    except:
        print("Creating experiment `{}`".format(experiment_name))
        return experiment.Experiment.create(
            experiment_name=experiment_name, sagemaker_boto_client=client)



import sagemaker

import pprint
from aws_sagemaker_remote.util.fields import get_field


def training_describe(job_name, client, field=None):
    description = training_describe_get(job_name=job_name, client=client)
    description = get_field(description, field)
    return description


def training_describe_get(client, job_name):
    response = client.describe_training_job(
        TrainingJobName=job_name
    )
    return response


if __name__ == '__main__':
    session = sagemaker.Session()
    client = session.boto_session.client('sagemaker')
    training_describe(
        client=client, job_name='mnist-demo-2020-10-06-04-32-42-393')



from ..session import sagemaker_session
from .config import SageMakerTrainingConfig
import os
from sagemaker.pytorch import PyTorch
from .channels import standardize_channels, upload_local_channels
from sagemaker.utils import name_from_base
from .iam import ensure_training_role
from .experiment import ensure_experiment
from ..git import git_get_tags
from ..tags import make_tags
from ..s3 import get_file_type, FileType


def sagemaker_training_run(
    args,
    config: SageMakerTrainingConfig,
    metrics=None
):
    if os.getenv('SM_TRAINING_ENV', None):
        warnings.warn(
            "Trying to start a SageMaker container from a SageMaker container. Possible loop detected.")
    if metrics is None:
        metrics = {}
    session = sagemaker_session(
        profile_name=args.sagemaker_profile
    )
    script = args.sagemaker_script
    script = os.path.abspath(script)
    source = args.sagemaker_source
    if not source:
        source = os.path.dirname(script)
    if not script.startswith(source):
        raise ValueError("script=[{}] must be in source=[{}]")
    entry_point = script[len(source)+1:]
    entry_point = entry_point.replace("\\", "/")
    metric_definitions = [
        {'Name': k, 'Regex': v}
        for k, v in metrics.items()
    ]
    dependencies = [getattr(args, k) for k in config.dependencies.keys()]

    # checkpoint_local_path='/opt/ml/checkpoints/'
    bucket = session.default_bucket()
    if args.sagemaker_job_name and args.sagemaker_job_name.strip():
        job_name = args.sagemaker_job_name
    else:
        job_name = name_from_base(args.sagemaker_base_job_name)
    tags = git_get_tags(script)
    tags["Source"] = 'aws-sagemaker-remote'
    tags["JobName"] = job_name
    tags["BaseJobName"] = args.sagemaker_base_job_name
    tags = make_tags(tags)
    #checkpoint_s3_uri = 's3://{}/{}/checkpoints'.format(bucket, job_name)
    input_prefix = "s3://{}/{}/inputs".format(bucket, job_name)
    iam = session.boto_session.client('iam')
    training_role = ensure_training_role(
        iam=iam, role_name=args.sagemaker_training_role)
    hyperparameters = {k.replace('_', '-'): str(v)
                       for k, v in vars(args).items() if v is not None and len(str(v)) > 0}
    hyperparameters['sagemaker-run'] = 'False'
    if args.sagemaker_checkpoint_s3 and args.sagemaker_checkpoint_s3 != 'default':
        if not args.sagemaker_checkpoint_s3.startswith('s3://'):
            raise ValueError(
                "--sagemaker-checkpoint-s3 must be an S3 URI (s3://...) or \"default\"")
        checkpoint_s3 = args.sagemaker_checkpoint_s3
    else:
        checkpoint_s3 = "s3://{}/{}/checkpoints".format(bucket, job_name)
    hyperparameters['checkpoint-dir'] = args.sagemaker_checkpoint_container
    if 'sagemaker-job-name' in hyperparameters:
        del hyperparameters['sagemaker-job-name']

    channels = config.inputs
    channels = {k: getattr(args, k) for k in channels.keys()}
    channels = {k: v for k, v in channels.items() if v} # and config.inputs[k].required == False}
    #for k,v in channels.items():
    #    if not v:
    #        raise ValueError("Channel [{}] is empty and required".format(k))
    channels = standardize_channels(channels=channels)
    channels = upload_local_channels(
        channels=channels, session=session, prefix=input_prefix)

    s3 = session.boto_session.client('s3')
    for k, v in channels.items():
        key = '{}-suffix'.format(k.replace('_', '-'))
        fileType = get_file_type(v, s3=s3)
        if fileType == FileType.FILE:
            hyperparameters[key] = os.path.basename(v)
        elif fileType == FileType.FOLDER:
            if key in hyperparameters:
                del hyperparameters[key]
        else:
            raise ValueError()
    print("Hyperparameters: {}".format(hyperparameters))

    if not channels:
        channels = None
    #env = config.env

    estimator = PyTorch(
        sagemaker_session=session,
        base_job_name=args.sagemaker_base_job_name,
        entry_point=entry_point,
        source_dir=source,
        role=training_role,
        instance_type=args.sagemaker_training_instance,
        image_uri=args.sagemaker_training_image,
        instance_count=1,
        framework_version='1.5.0',
        # hyperparameters=hyperparameters_from_argparse(vars(args)),
        metric_definitions=metric_definitions,
        dependencies=dependencies,
        checkpoint_s3_uri=checkpoint_s3,
        checkpoint_local_path=args.sagemaker_checkpoint_container,
        use_spot_instances=args.sagemaker_spot_instances,
        hyperparameters=hyperparameters,
        volume_size=args.sagemaker_volume_size,
        tags=tags,
        max_wait=args.sagemaker_max_wait if args.sagemaker_spot_instances else None,
        max_run=args.sagemaker_max_run
    )

    if args.sagemaker_experiment_name:
        sagemaker_client = session.boto_session.client('sagemaker')
        ensure_experiment(client=sagemaker_client,
                          experiment_name=args.sagemaker_experiment_name)
        experiment_config = {
            "ExperimentName": args.sagemaker_experiment_name
        }
        if args.sagemaker_trial_name:
            experiment_config["TrialName"] = args.sagemaker_trial_name
    else:
        if args.sagemaker_trial_name:
            raise ValueError(
                "If `sagemaker_trial_name` is provided, `sagemaker_experiment_name` must be provided as well")
        experiment_config = None

    estimator.fit(channels, job_name=job_name,
                  wait=args.sagemaker_wait, experiment_config=experiment_config)
    # todo:
    # use_spot_instances
    # experiment_config (dict[str, str]): Experiment management configuration.
    #            Dictionary contains three optional keys,
    #            'ExperimentName', 'TrialName', and 'TrialComponentDisplayName'.
    return estimator




from argparse import ArgumentParser, Namespace


class Command(object):
    def __init__(self, help=None):
        self.help = help

    def configure(self, parser: ArgumentParser):
        pass

    def run(self, args):
        pass


def run_command(command: Command, description=None):
    parser = ArgumentParser(description=description or command.help)
    command.configure(parser)
    args = parser.parse_args()
    return command.run(args)


def commands_parser(
    commands,
    description=None,
    parser=None,
    dest='command'
):
    if parser is None:
        description = description or "Experiment CLI"
        parser = ArgumentParser(description=description)
    parser_commands = parser.add_subparsers(
        title='command',
        dest=dest,
        help='Command to execute',
        #required=True todo: py 3.7 only
    )
    for k, command in commands.items():
        parser_command = parser_commands.add_parser(
            name=k, help=command.help
        )
        command.configure(parser_command)
    return parser


def handle_commands(commands, args):
    command = commands[args.command]
    vargs = vars(args)
    del vargs['command']
    args = Namespace(**vargs)
    return command.run(args)


def run_commands(commands, description=None, argv=None,dry_run=False):
    parser = commands_parser(
        commands=commands,
        description=description
    )
    if dry_run:
        return parser
    else:
        args = parser.parse_args(args=argv)
        return handle_commands(
            commands=commands,
            args=args
        )



from botocore.exceptions import ClientError
from ..iam import SAGEMAKER_FULL_ACCESS, ensure_role

TRAINING_ROLE = {
    'description': 'Role for SageMaker training (aws-sagemaker-remote)',
    'policies': [SAGEMAKER_FULL_ACCESS],
    'trust': """
{
  "Version": "2008-10-17",
  "Statement": [
    {
      "Sid": "",
      "Effect": "Allow",
      "Principal": {
        "Service": "sagemaker.amazonaws.com"
      },
      "Action": "sts:AssumeRole"
    }
  ]
}
"""
}


def ensure_training_role(iam, role_name):
    return ensure_role(iam=iam, role_name=role_name, **TRAINING_ROLE)



import re
MAX_LENGTH = 256
SUFFIX = '...'


def clip_tag(value, max_length=MAX_LENGTH, suffix=SUFFIX):
    if len(value) > max_length:
        return "{}{}".format(value[:max_length-len(SUFFIX)], suffix)
    else:
        return str(value)


def clean_tag(value:str):
    value = value.strip()
    value = re.sub("\\s+", "-", value)
    value = "".join(re.findall("[\\w\\=\\-\\.:_\\+@/]", value))
    return value


def make_tags(tags):
    tags = [
        {
            "Key": k,
            "Value": clip_tag(clean_tag(v))
        }
        for k, v in tags.items() if v
    ]
    tags = [tag for tag in tags if tag['Value']]
    return tags

if __name__=='__main__':
    print(clean_tag("[a]sd a/sd-_/+@"))



import json
def output_fn_json(prediction, content_type):

    if content_type == "application/json":
        return json.dumps(prediction), content_type
    else:
        raise ValueError("Unsupported content type: {}".format(content_type))


import boto3
from urllib.request import urlparse
from botocore.exceptions import ClientError

class FileType(object):
    FILE = 'File'
    FOLDER = 'Folder'

def is_s3_folder(uri, s3):
    url = urlparse(uri)
    assert url.scheme == 's3'
    bucket = url.hostname
    key = url.path
    if key.startswith('/'):
        key = key[1:]
    if not key.endswith('/'):
        key += '/'
    try:
        response = s3.list_objects_v2(
            Bucket=bucket,
            #Delimiter='string',
            #EncodingType='url',
            MaxKeys=1,
            Prefix=key,
            #ContinuationToken='string',
            #FetchOwner=True|False,
            #StartAfter='string',
            #RequestPayer='requester',
            #ExpectedBucketOwner='string'
        )
        if response['KeyCount']:
            return True
        else:
            return False
    except ClientError:
        return False

def is_s3_file(uri, s3):
    url = urlparse(uri)
    assert url.scheme == 's3'
    bucket = url.hostname
    key = url.path
    if key.startswith('/'):
        key = key[1:]
    try:
        response = s3.head_object(
            Bucket=bucket,
            Key=key)
        return response
    except ClientError:
        return None


def get_file_type(uri, s3):
    if is_s3_file(uri, s3):
        return FileType.FILE
    elif is_s3_folder(uri, s3):
        return FileType.FOLDER
    else:
        raise ValueError("Cannot determine if URI [{}] is file or folder. Check your permissions, your AWS profile connection, and that URI exists.".format(uri))

if __name__ == '__main__':
    session = boto3.Session()
    s3 = session.client('s3')
    uri = 's3://sagemaker-us-east-1-683880991063/demo-training-inputs-2020-10-01-15-49-26-167/inputs/test_folder/test_file.txt'
    print(get_file_type(uri, s3))
    uri = 's3://sagemaker-us-east-1-683880991063/demo-training-inputs-2020-10-01-15-49-26-167/inputs/test_folder'
    print(get_file_type(uri, s3))
    uri = 's3://sagemaker-us-east-1-683880991063/demo-training-inputs-2020-10-01-15-49-26-167/inputs/test_folder/'
    print(get_file_type(uri, s3))
    uri = 's3://sagemaker-us-east-1-683880991063/demo-training-inputs-2020-10-01-15-49-26-167/inputs/test_fold'
    print(get_file_type(uri, s3))



import argparse
from distutils.util import strtobool
import os
from ..args import variable_to_argparse, bool_argument, sagemaker_profile_args, PROFILE
from .config import SageMakerTrainingConfig
import json
import inspect
import types

CHANNEL_HELP = """Input channel [{channel}].
Set to local path and it will be uploaded to S3 and downloaded to SageMaker.
Set to S3 path and it will be downloaded to SageMaker.
(default: [{default}])
"""
TRAINING_IMAGE = '683880991063.dkr.ecr.us-east-1.amazonaws.com/columbo-sagemaker-training:latest'
TRAINING_INSTANCE = 'ml.m5.large'
TRAINING_ROLE = 'aws-sagemaker-remote-training-role'
BASE_JOB_NAME = 'training-job'
CHECKPOINT_LOCAL_PATH = '/opt/ml/checkpoints'

"""
def sagemaker_env_args(config):
    args = {}
    output_dir = os.environ.get('SM_OUTPUT_DIR', None)
    if output_dir:
        args['output_dir'] = output_dir
    model_dir = os.environ.get('SM_MODEL_DIR', None)
    if model_dir:
        args['model_dir'] = model_dir
    for channel in config.inputs.keys():
        env_key = 'SM_CHANNEL_{}'.format(channel.upper())
        channel_dir = os.environ.get(env_key, None)
        if channel_dir:
            args[channel] = channel_dir
    return args
"""

def sagemaker_env_arg():
    """
    Check for ``SM_TRAINING_ENV`` environment variable and return object if it exists
    """
    sm_env = os.getenv('SM_TRAINING_ENV', None)
    if sm_env:
        data = json.loads(sm_env)
        return data
    else:
        return None

def sagemaker_env_args(args: argparse.Namespace, config: SageMakerTrainingConfig):
    """
    Check for ``SM_TRAINING_ENV`` environment variable and use it to override arguments.
    """
    kwargs = vars(args)
    data = sagemaker_env_arg()
    if data:
        print("Detected SageMaker environment")
        ci = data.get('channel_input_dirs', None)
        if ci:
            for k in config.inputs.keys():
                v = ci.get(k, None)
                if v:
                    suffix = kwargs.get('{}_suffix'.format(k), None)
                    if suffix:
                        v = os.path.join(v, suffix)
                    print("SageMaker input [{}]: [{}]".format(k, v))
                    kwargs[k] = v
                else:
                    kwargs[k] = None
        else:
            for k in config.inputs.keys():
                kwargs[k] = None
        output_dir = data.get('output_data_dir', None)
        if output_dir:
            kwargs['output_dir'] = output_dir
            print("SageMaker output_dir: [{}]".format(output_dir))
        model_dir = data.get('model_dir', None)
        if model_dir:
            kwargs['model_dir'] = model_dir
            print("SageMaker model_dir: [{}]".format(model_dir))
        job_name = data.get('job_name', None)
        if job_name:
            kwargs['sagemaker_job_name'] = job_name
            print("SageMaker job_name: [{}]".format(job_name))
        kwargs['sagemaker_run'] = False
        kwargs['is_sagemaker'] = True
    else:
        kwargs['is_sagemaker'] = False
    new_args = argparse.Namespace(
        **kwargs
    )
    #print("Original args: {}".format(args))
    #print("Updated args: {}".format(new_args))
    return new_args


def sagemaker_training_args(
    parser: argparse.ArgumentParser,
    script,
    source='',
    base_job_name=BASE_JOB_NAME,
    job_name='',
    profile=PROFILE,
    run=False,
    wait=True,
    inputs=None,
    dependencies=None,
    additional_arguments=None,
    argparse_callback=None,
    model_dir='output',
    output_dir='output',
    checkpoint_dir='output',
    checkpoint_s3='default',
    checkpoint_container=CHECKPOINT_LOCAL_PATH,
    training_image=TRAINING_IMAGE,
    training_instance=TRAINING_INSTANCE,
    training_role=TRAINING_ROLE,
    enable_sagemaker=True,
    experiment_name=None,
    trial_name=None,
    spot_instances=False,
    volume_size=30,
    max_run=60*60*12,
    max_wait=60*60*24,
    env=None
):
    r"""
    Configure ``argparse.ArgumentParser`` for training scripts.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        Parser to configure
    script : str
        Path to script file to execute.
        Set default for ``--sagemaker-script``
    source : str, optional
        Path of source directory to upload.
        Must include ``script`` path.
        Defaults to directory containing ``script`` if not provided.
    base_job_name : str, optional
        Job name will be generated from ``base_job_name`` and a timestamp if ``job_name`` is not provided.
        Set default for ``--sagemaker-base-job-name``.
    job_name : str, optional
        Job name is used for tracking and organization. 
        Generated from ``base_job_name`` if not provided.
        Use ``base_job_name`` and leave ``job_name`` blank for most use-cases.
        Set default for ``--sagemaker-job-name``.
    profile : str, optional
        AWS profile to use for session. 
        Set default for ``--sagemaker-profile``.
    run : bool, optional
        Run on SageMaker. 
        Set default for ``--sagemaker-run``.
    wait : bool, optional
        Wait for SageMaker processing to complete. 
        Set default for ``--sagemaker-wait``.
    inputs : dict(str,str), optional
        Dictionary of input arguments.
        For eack key and value, create an argument ``--key`` that defaults to value.
        * Running locally, input arguments are unmodified.
        * Running remotely, inputs are set to appropriate SageMaker mount points. Local inputs are uploaded automatically.
    dependencies : dict(str, str)
        Dictionary of modules.
        For eack key and value, create an argument ``--module-key`` that defaults to value. 
        This controls the path of a dependency of your code.
        The files at the given path will be uploaded to S3, downloaded to SageMaker, and put on PYTHONPATH.    
    additional_arguments: list, optional
        List of tuple of positional args and keyword args for ``argparse.ArgumentParser.add_argument``.
        Use to add additional arguments to the script.
    argparse_callback: function, optional
        Function accepting one argument ``parser:argparse.ArgumentParser`` that adds additional arguments.
        Use to add additional arguments to the script.
    model_dir: string, optional
        Directory to save trained inference model.
        Set default for ``--model-dir``.
    output_dir: string, optional
        Directory to save outputs (images, logs, etc.).
        Set default for ``--output-dir``.
    checkpoint_dir: string, optional
        Directory to save checkpoints for saving and resuming training.
        Set default for ``--checkpoint-dir``.
    checkpoint_s3: string, optional
        S3 storage for checkpoints for saving and resuming training or "default".
        Set default for ``--sagemaker-checkpoint-s3``.
    checkpoint_container: string, optional
        Local directory for checkpoints when running remotely.
        Set default for ``--sagemaker-checkpoint-container``.
    training_image : str, optional
        URI of ECR or DockerHub Docker image to use for training. 
        Set default for ``--sagemaker-training-image``.
    training_instance : str, optional
        Type of instance to use for training (e.g., ``ml.t3.medium``). 
        Set default for ``--sagemaker-training-instance``.
    training_role : str, optional
        AWS IAM role name to use for training. Will be created if it does not exist. 
        Set default for ``--sagemaker-training-role``.
    experiment_name : str, optional
        Name of experiment. Required if ``trial_name`` is provided.
        Set default for ``--sagemaker-experiment-name``.
    trial_name : str, optional
        Name of trial within experiment. 
        Set default for ``--sagemaker-trial-name``.
    enable_sagemaker : bool, optional
        * True: Include SageMaker command-line options.
        * False: Only include local command-line options
    max_run : int, optional
        Maximum training time in seconds.
    max_wait : int, optional
        Maximum time to wait for a spot instance in seconds.
    """
    if isinstance(script, types.FunctionType):
        script = inspect.getfile(script)
    config = SageMakerTrainingConfig(inputs=inputs, dependencies=dependencies, env=env)
    if enable_sagemaker:
        group = parser.add_argument_group(
            title='SageMaker', description='SageMaker options'
        )
        sagemaker_profile_args(parser=group, profile=profile)
        bool_argument(group, '--sagemaker-run', default=run,
                      help="Run training on SageMaker (yes/no default={})".format(run))
        bool_argument(group, '--sagemaker-wait', default=wait,
                      help="Wait for SageMaker training to complete and tail logs files (yes/no default={})".format(wait))
        bool_argument(group, '--sagemaker-spot-instances', default=spot_instances,
                      help="Use spot instances for training (yes/no default={})".format(spot_instances))
        group.add_argument(
            '--sagemaker-script',
            default=script,
            help='Script to run on SageMaker. (default: [{}])'.format(script))
        group.add_argument(
            '--sagemaker-source',
            default=source,
            help='Source to upload to SageMaker. '
            'Must contain script. '
            'If blank, default to directory containing script. '
            '(default: [{}])'.format(source))
        group.add_argument(
            '--sagemaker-training-instance',
            default=training_instance,
            help='Instance type for training')
        group.add_argument(
            '--sagemaker-training-image',
            default=training_image,
            help='Docker image for training')
        group.add_argument(
            '--sagemaker-training-role',
            default=training_role,
            help='Docker image for training')
        group.add_argument(
            '--sagemaker-base-job-name',
            default=base_job_name,
            help='Base job name for tracking and organization on S3.'
            ' A job name will be generated from the base job name unless a job name is specified.')
        group.add_argument(
            '--sagemaker-job-name',
            default=job_name,
            help='Job name for tracking. Use --base-job-name instead and a job name will be automatically generated with a timestamp.')
        group.add_argument(
            '--sagemaker-experiment-name',
            default=experiment_name,
            help='Name of experiment in SageMaker tracking.')
        group.add_argument(
            '--sagemaker-trial-name',
            default=trial_name,
            help='Name of experiment trial in SageMaker tracking.')
        group.add_argument(
            '--sagemaker-volume-size',
            type=int,
            default=volume_size,
            help='Volume size in GB.')
        group.add_argument(
            '--sagemaker-max-run',
            type=int,
            default=max_run,
            help='Maximum runtime in seconds.')
        group.add_argument(
            '--sagemaker-max-wait',
            type=int,
            default=max_wait,
            help='Maximum time to wait for spot instances in seconds.')
        sagemaker_training_dependency_args(
            parser=parser, dependencies=config.dependencies)
    sagemaker_training_model_args(parser=parser, model_dir=model_dir)
    sagemaker_training_output_args(parser=parser, output_dir=output_dir)
    sagemaker_training_checkpoint_args(
        parser=parser,
        checkpoint_dir=checkpoint_dir,
        checkpoint_s3=checkpoint_s3,
        checkpoint_container=checkpoint_container,
        enable_sagemaker=enable_sagemaker)
    sagemaker_training_channel_args(parser=parser, inputs=inputs)
    if additional_arguments:
        for args, kwargs in additional_arguments:
            parser.add_argument(*args, **kwargs)
    if argparse_callback:
        argparse_callback(parser)
    return config


def sagemaker_training_output_args(parser: argparse.ArgumentParser, output_dir):
    output_dir = os.environ.get('SM_OUTPUT_DIR', output_dir)
    parser.add_argument('--output-dir', type=str,
                        default=output_dir,
                        help='Directory for logs, images, or other output files (default: "{}")'.format(output_dir))


def sagemaker_training_checkpoint_args(
        parser: argparse.ArgumentParser, checkpoint_dir,
        checkpoint_s3='default', checkpoint_container=CHECKPOINT_LOCAL_PATH, enable_sagemaker=True):

    group = parser.add_argument_group(
        title='Checkpoints',
        description='Checkpointing options'
    )
    group.add_argument('--checkpoint-dir', type=str,
                       default=checkpoint_dir,
                       help='Local directory to store checkpoints for resuming training (default: "{}")'.format(checkpoint_dir))
    if enable_sagemaker:
        group.add_argument('--sagemaker-checkpoint-s3', type=str,
                           default=checkpoint_s3,
                           help='Location to store checkpoints on S3 or "default" (default: "{}")'.format(checkpoint_s3))
        group.add_argument('--sagemaker-checkpoint-container', type=str,
                           default=checkpoint_container,
                           help='Location to store checkpoints on container (default: "{}")'.format(checkpoint_container))


def sagemaker_training_dependency_args(parser: argparse.ArgumentParser, dependencies):
    if dependencies:
        group = parser.add_argument_group(
            title='Dependencies',
            description='Dependencies to upload to SageMaker'
        )
        for k, v in dependencies.items():
            flag = variable_to_argparse(k)
            group.add_argument(flag, type=str,
                               default=k,
                               help='Directory for dependency [{}] (default: "{}")'.format(k, v))


def sagemaker_training_channel_args(parser: argparse.ArgumentParser, inputs):
    if inputs:
        group = parser.add_argument_group(
            title='Inputs',
            description='Inputs (local or S3)'
        )
        for channel, default in inputs.items():
            flag = variable_to_argparse(channel)
            env_key = 'SM_CHANNEL_{}'.format(channel.upper())
            if env_key in os.environ:
                default = os.environ[env_key]
            group.add_argument(
                flag, type=str,  default=default,
                help=CHANNEL_HELP.format(channel=channel, default=default))
            suffix_flag = variable_to_argparse("{}_suffix".format(channel))
            group.add_argument(
                suffix_flag, help=argparse.SUPPRESS, default="", type=str)


def sagemaker_training_model_args(parser: argparse.ArgumentParser,
                                  model_dir='model'
                                  ):
    # Overwrite default if running on SageMaker
    model_dir = os.environ.get('SM_MODEL_DIR', model_dir)
    parser.add_argument(
        '--model-dir', type=str,
        default=model_dir,
        help='Directory to save final model (default: {})'.format(model_dir))


def sagemaker_training_parser_for_docs():
    parser = argparse.ArgumentParser()
    script = 'script.py'
    sagemaker_training_args(
        parser=parser, script=script,
        inputs={
            'input': 'path/to/input'
        },
        dependencies={
            'my_module': 'path/to/my_module'
        })
    return parser



import argparse
from distutils.util import strtobool
import os
from urllib.parse import urlparse
from urllib.request import url2pathname

PROFILE = 'default'


def get_local_path(path):
    if not path:
        return None
    elif os.path.exists(path):
        return os.path.abspath(path)
    else:
        try:
            url = urlparse(path)
            if url.scheme in ['file', '']:
                path = url2pathname(url.path)
                return os.path.abspath(path)
            elif url.scheme in ['s3']:
                return None
            else:
                raise ValueError("Unknown scheme or path does not exist: [{}]".format(path))
        except:
            pass
        return None

class PathArgument(object):
    def __init__(
        self,
        local=None,
        remote='default',
        optional=False,
        mode=None
    ):
        self.local = local
        self.remote = remote or 'default'
        self.optional = optional
        self.mode = mode
        if self.mode:
            assert self.mode in ['Pipe', 'File','EndOfJob', 'Continuous']
        #if self.optional and self.local:
        #    print("Optional inputs must default to nothing")

OPTIONAL = PathArgument(optional=True)

def convert_path_argument(param, cls=PathArgument):
    if isinstance(param, cls):
        return param
    elif isinstance(param, str):
        return cls(param)
    else:
        raise ValueError(
            "Unexpected path type [{}]: [{}]".format(type(param), param))


def convert_path_arguments(params, cls=PathArgument):
    if params:
        return {
            k: convert_path_argument(v, cls=cls) for k, v in params.items()
        }
    else:
        return {}


def variable_to_argparse(variable):
    return "--{}".format(variable.replace("_", "-"))


def argparse_to_variable(flag):
    assert flag.startswith('--')
    return flag[2:].replace('-', '_')


def strtobool_type(v):
    return bool(strtobool(v))


def bool_argument(
    parser: argparse.ArgumentParser,
    *args,
    **kwargs
):
    parser.add_argument(*args, **kwargs, type=strtobool_type,
                        const=True, nargs="?")


def sagemaker_profile_args(parser, profile=PROFILE):
    parser.add_argument('--sagemaker-profile', default=profile,
                        help='AWS profile for SageMaker session (default: [{}])'.format(profile))



import click
from aws_sagemaker_remote.util.training import training_describe
from aws_sagemaker_remote.util.processing import processing_describe
from aws_sagemaker_remote.inference.model import model_create, model_delete, model_describe
from aws_sagemaker_remote.inference.endpoint import endpoint_create, endpoint_delete, endpoint_describe, endpoint_invoke
from aws_sagemaker_remote.inference.endpoint_config import endpoint_config_create, endpoint_config_delete, endpoint_config_describe
import sagemaker
import boto3
from aws_sagemaker_remote.util.upload import upload
import logging

logging.getLogger('boto3').setLevel(logging.CRITICAL)
logging.getLogger('botocore').setLevel(logging.CRITICAL)

current_profile = None


@click.group()
@click.option('--profile', help='AWS profile. Run `aws-configure` to configure a profile.', default='default')
def cli(profile='default'):
    global current_profile
    current_profile = profile


@cli.command(name='upload')
@click.argument('src')
@click.argument('dst')
@click.option('--gz/--no-gz', default=False)
def upload_cli(src, dst, gz):
    session = boto3.Session(profile_name=current_profile)
    session = sagemaker.Session(session)
    upload(src=src, dst=dst, gz=gz, session=session)


@cli.group()
def processing():
    pass
    
@processing.command(name='describe')
@click.argument(
    'name'
)
@click.argument(
    'field',
    required=False)
def processing_describe_cli(name, field):
    """
    Describe training job NAME. 

    * Print full JSON if FIELD is not specified.

    * Print only FIELD if specified (e.g., ModelArtifacts.S3ModelArtifacts or LastModifiedTime).
    """
    session = boto3.Session(profile_name=current_profile)
    client = session.client('sagemaker')
    description = processing_describe(
        job_name=name,
        field=field,
        client=client
    )
    print(description)


@cli.group()
def training():
    pass


@training.command(name='describe')
@click.argument(
    'name'
)
@click.argument(
    'field',
    required=False)
def training_describe_cli(name, field):
    """
    Describe training job NAME. 

    * Print full JSON if FIELD is not specified.

    * Print only FIELD if specified (e.g., ModelArtifacts.S3ModelArtifacts or LastModifiedTime).
    """
    session = boto3.Session(profile_name=current_profile)
    client = session.client('sagemaker')
    description = training_describe(
        job_name=name,
        field=field,
        client=client
    )
    print(description)

# Model CLI


@cli.group()
def model():
    """
    Model CLI
    """
    pass


@model.command(name='create')
@click.option('--job', help='Job name', type=str, default=None)
@click.option('--model-artifact', help='Model artifact (S3 URI). Relative path assumes default bucket', type=str, default=None)
@click.option('--name', help='Model name', type=str, default=None)
@click.option('--inference-image', help='ECR Docker URI for inference', type=str,
              default='683880991063.dkr.ecr.us-east-1.amazonaws.com/columbo-sagemaker-inference:latest')
@click.option('--role', help='SageMaker inference role name', type=str,
              default='aws-sagemaker-remote-inference-role')
@click.option('--force/--no-force', default=False)
def model_create_cli(job, model_artifact, name, inference_image, role, force):
    session = boto3.Session(profile_name=current_profile)
    session = sagemaker.Session(session)
    model_create(
        job=job,
        model_artifact=model_artifact,
        name=name,
        session=session,
        inference_image=inference_image,
        role=role,
        force=force)


@model.command(name='delete')
@click.argument('name')
def model_delete_cli(name):
    """
    Delete SageMaker model NAME
    """
    session = boto3.Session(profile_name=current_profile)
    client = session.client('sagemaker')
    model_delete(name=name, client=client)


@model.command(name='describe')
@click.argument("name")
@click.argument("field", required=False, default=None)
def model_describe_cli(name, field):
    """
    Describe endpoint configuration NAME
    """
    session = boto3.Session(profile_name=current_profile)
    client = session.client('sagemaker')
    description = model_describe(
        name=name,
        client=client,
        field=field
    )
    print(description)

# Endpoint config CLI


@cli.group()
def endpoint_config():
    """
    Endpoint config CLI
    """
    pass


@endpoint_config.command(name='create')
@click.option('--name', type=str)
@click.option('--model', multiple=True)
@click.option(
    '--instance-type',
    help='SageMaker instance type', type=str,
    default='ml.t2.medium')
@click.option('--force/--no-force', default=False)
def endpoint_config_create_cli(name, model, instance_type, force):
    session = boto3.Session(profile_name=current_profile)
    session = sagemaker.Session(session)
    endpoint_config_create(
        name=name,
        model=model,
        instance_type=instance_type,
        force=force,
        session=session)


@endpoint_config.command(name='describe')
@click.argument("name")
@click.argument("field", required=False, default=None)
def endpoint_config_describe_cli(name, field):
    """
    Describe endpoint configuration NAME
    """
    session = boto3.Session(profile_name=current_profile)
    client = session.client('sagemaker')
    description = endpoint_config_describe(
        name=name,
        client=client,
        field=field
    )
    print(description)


@endpoint_config.command(name='delete')
@click.argument("name")
def endpoint_config_delete_cli(name):
    """
    Delete endpoint configuration NAME
    """
    session = boto3.Session(profile_name=current_profile)
    client = session.client('sagemaker')
    endpoint_config_delete(
        name=name,
        client=client
    )

# Endpoint CLI


@cli.group()
def endpoint():
    """
    Endpoint CLI
    """
    pass


@endpoint.command(name='create')
@click.option('--name', type=str, default=None, help='Name of endpoint')
@click.option('--config', type=str, default=None, help='Name of endpoint config')
@click.option('--force/--no-force', default=False, help='Overwrite existing endpoint')
def endpoint_create_cli(name, config, force):
    session = boto3.Session(profile_name=current_profile)
    client = session.client('sagemaker')
    endpoint_create(
        name=name, config=config, force=force,
        client=client)


@endpoint.command(name='delete')
@click.argument('name', type=str)
def endpoint_delete_cli(name):
    session = boto3.Session(profile_name=current_profile)
    client = session.client('sagemaker')
    endpoint_delete(
        name=name,
        client=client)


@endpoint.command(name='describe')
@click.argument('name', type=str)
@click.argument('field', type=str, default=None, required=False)
def endpoint_describe_cli(name, field):
    session = boto3.Session(profile_name=current_profile)
    client = session.client('sagemaker')
    description = endpoint_describe(
        name=name,
        client=client,
        field=field)
    print(description)


@endpoint.command(name='invoke')
@click.option('--name', type=str, default=None)
@click.option('--model', type=str, default=None)
@click.option('--variant', type=str, default=None)
@click.option('--input', type=str, default=None)
@click.option('--output', type=str, default=None)
@click.option('--input-type', type=str, default=None)
@click.option('--output-type', type=str, default=None)
@click.option('--model-dir', type=str, default=None)
def endpoint_invoke_cli(name, model, variant, input, output, input_type, output_type, model_dir):
    session = boto3.Session(profile_name=current_profile)
    runtime_client = session.client('sagemaker-runtime')
    result = endpoint_invoke(
        name=name,
        model=model,
        variant=variant,
        input=input,
        output=output,
        input_type=input_type,
        output_type=output_type,
        model_dir=model_dir,
        runtime_client=runtime_client)
    if result:
        print(result)


if __name__ == '__main__':
    cli()



from setuptools import find_packages, setup
import os
with open(os.path.abspath(os.path.join(__file__, '../README.rst')), encoding='utf-8') as f:
    long_description = f.read()
setup(name='aws-sagemaker-remote',
      version='0.0.33',
      author='Ben Striner',
      author_email="bstriner@gmail.com",
      url='https://github.com/bstriner/aws-sagemaker-remote',
      install_requires=[
          'sagemaker',
          'sagemaker-experiments',
          'click'
      ],
      entry_points={
          'console_scripts':[
              'aws-sagemaker-remote=aws_sagemaker_remote.cli:cli'
          ]
      },
      description="Simplify running processing and training remotely on AWS SageMaker",
      packages=find_packages(),
      long_description=long_description,
      long_description_content_type='text/x-rst')




def get_field(data, field):
    if field:
        path = field.split('.')
        for p in path:
            if p not in data:
                raise ValueError("Field does not exist: [{}]".format(field))
            data = data[p]
    return data



from ..modules import module_path


class SageMakerTrainingConfig(object):
    def __init__(self, inputs=None, dependencies=None, env=None):
        self.inputs = inputs or {}
        self.dependencies = dependencies or {}
        self.dependencies = {k: module_path(v)
                             for k, v in self.dependencies.items()}
        self.env=env or {}



import sagemaker
import click
import mimetypes
import os
from botocore.exceptions import ClientError
from aws_sagemaker_remote.util.fields import get_field
from aws_sagemaker_remote.inference.local import inference_local
import shutil


def endpoint_describe(name, client, field=None):
    description = client.describe_endpoint(EndpointName=name)
    description = get_field(description, field)
    return description


def endpoint_exists(name, client):
    try:
        endpoint_describe(name, client)
        return True
    except ClientError:
        return False


def endpoint_delete(name, client):
    client.delete_endpoint(EndpointName=name)


def endpoint_create(config, name, client, force):
    if not config:
        raise click.UsageError('Must specify endpoint config')
    if not name:
        name = config
    print("Creating endpoint [{}] from endpoint config [{}]".format(
        name, config
    ))
    if endpoint_exists(name=name, client=client):
        if force:
            print("Deleting existing endpoint")
            endpoint_delete(name=name, client=client)
            #todo: need to wait for delete to complete
        else:
            raise click.UsageError(
                'Specify force to overwrite existing endpoint')
    response = client.create_endpoint(
        EndpointName=name,
        EndpointConfigName=config)
    arn = response['EndpointArn']
    print("Created endpoint [{}]".format(arn))


def endpoint_invoke(model_dir, name, model, variant, input, output, input_type, output_type, runtime_client):
    if not input:
        # todo: pipe and cli arguments
        raise click.UsageError("input is required")
    if not input_type:
        input_type,_ = mimetypes.guess_type(input)
        if not input_type:
            raise click.UsageError(
                'Cannot guess input type, specify input_type')
    if not output_type:
        output_type = 'application/json'
    if model_dir:
        if name or model or variant:
            raise click.UsageError("If a local model_dir is specified, do not specify name, model or variant")
        output_data = inference_local(model_dir=model_dir, input=input,
        input_type=input_type,output_type=output_type, output=output)
    else:
        kwargs = {}
        if model:
            kwargs['TargetModel'] = model
        if model:
            kwargs['TargetVariant'] = variant
        with open(input, 'rb') as f:
            print("Invoking [{}] with data from [{}] ([{}]->[{}])".format(
                name,input, input_type, output_type
            ))
            response = runtime_client.invoke_endpoint(
                EndpointName=name,
                Body=f,
                ContentType=input_type,
                Accept=output_type,
                # CustomAttributes='string',
                **kwargs
                # TargetModel=model,
                # TargetVariant=variant
            )
        output_data = response['Body']
        if output:
            os.makedirs(os.path.dirname(os.path.abspath(output)), exist_ok=True)
            """
            if isinstance(result[0], str):
                mode = 'w'
            elif isinstance(result[0], bytes):
                mode='wb'
            else:
                raise ValueError("Unknown result type: {}".format(type(result[0])))
            with open(output, mode) as f:
                for chunk in result:
                    f.write(chunk)
            """

            with open(output, 'wb') as f:
                shutil.copyfileobj(output_data, f)
            print("Saved to {}".format(output))
            return None
        else:
            return output_data


if __name__ == '__main__':
    pass



import os
from argparse import ArgumentParser
from pathlib import Path

try:
    from boto3.session import Session as Boto3Session
    from sagemaker.session import Session as SagemakerSession
    from sagemaker.processing import ScriptProcessor, ProcessingInput, ProcessingOutput
except:
    pass
import json

import datetime

from ..session import sagemaker_session
from .iam import ensure_processing_role
from ..args import variable_to_argparse, get_local_path, PathArgument
from .args import PROCESSING_INSTANCE, PROCESSING_IMAGE, PROCESSING_JOB_NAME, PROCESSING_RUNTIME_SECONDS, INPUT_MOUNT, OUTPUT_MOUNT, MODULE_MOUNT
from .config import SageMakerProcessingConfig
from .args import sagemaker_processing_args
from ..git import git_get_tags
from ..tags import make_tags
from ..s3 import is_s3_file
PROCESSING_SCRIPT = os.path.abspath(os.path.join(__file__, '../processing.sh'))


def sagemaker_processing_run(args, config):
    script = args.sagemaker_script
    script = os.path.abspath(script)
    script = script.replace("\\", "/")

    session = sagemaker_session(profile_name=args.sagemaker_profile)

    inputs = {
        k: PathArgument(
            local=getattr(args, k),
            optional=v.optional,
            mode=getattr(args, "{}_mode".format(k) or v.mode or 'File')
        ) for k, v in config.inputs.items()
    }
    for k, v in inputs.items():
        if (not v.local) and (not v.optional):
            raise ValueError("Value required for input agument [{}]".format(k))
    inputs = {
        k: v for k, v in inputs.items() if v.local
    }
    outputs = {
        k: PathArgument(
            local=getattr(args, k),
            remote=getattr(args, "{}_s3".format(k)),
            optional=v.optional,
            mode=getattr(args, "{}_mode".format(k) or v.mode or 'EndOfJob')
        ) for k, v in config.outputs.items()
    }
    # for k, v in outputs.items():
    #    if (not v) and (not config.outputs[k].optional):
    #        raise ValueError(
    #            "Value required for output agument [{}_s3]".format(k))
    # outputs = {
    #    k: v for k, v in outputs.items() if v
    # }
    # todo: optional arguments
    dependencies = {
        k: getattr(args, k) for k in config.dependencies.keys()
    }
    tags = git_get_tags(script)
    process(
        inputs=inputs,
        outputs=outputs,
        dependencies=dependencies,
        session=session,
        role=args.sagemaker_role,
        script=script,
        image=args.sagemaker_image,
        instance=args.sagemaker_instance,
        base_job_name=args.sagemaker_base_job_name,
        job_name=args.sagemaker_job_name,
        volume_size=args.sagemaker_volume_size,
        python=args.sagemaker_python,
        runtime_seconds=args.sagemaker_runtime_seconds,
        output_mount=args.sagemaker_output_mount if hasattr(
            args, 'sagemaker_output_mount') else None,
        input_mount=args.sagemaker_input_mount if hasattr(
            args, 'sagemaker_input_mount') else None,
        module_mount=args.sagemaker_module_mount if hasattr(
            args, 'sagemaker_module_mount') else None,
        arguments=make_arguments(args=args, config=config),
        requirements=args.sagemaker_requirements,
        configuration_script=args.sagemaker_configuration_script,
        configuration_command=args.sagemaker_configuration_command,
        wait=args.sagemaker_wait,
        tags=tags,
        output_json=args.sagemaker_output_json,
        env=config.env
    )


def make_arguments(args, config: SageMakerProcessingConfig):
    vargs = vars(args)
    # , 'sagemaker_role', 'sagemaker_profile']
    to_del = ['sagemaker_run', 'sagemaker_job_name']
    to_del.extend(config.inputs.keys())
    to_del.extend(config.outputs.keys())
    to_del.extend("{}_s3" for k in config.outputs.keys())
    to_del.extend(config.dependencies.keys())
    for k in to_del:
        if k in vargs:
            del vargs[k]
    vargs['sagemaker_run'] = False
    return vargs


def sagemaker_arguments(vargs):
    arguments = []
    for k, v in vargs.items():
        if v is not None:
            v = str(v)
            if len(v) > 0:
                arguments.extend(
                    [variable_to_argparse(k), str(v)]
                )
    return arguments


def ensure_eol(file):
    """
    Ensure that file has Linux line endings. Convert if it doesn't.
    """
    if b"\r\n" in open(file, 'rb').read():
        print("DOS line endings found in {}. Attempting conversion.".format(file))
        with open(file, 'U') as infile:
            text = infile.read()
        with open(file, 'w', newline='\n') as outfile:
            outfile.write(text)


def make_processing_input(mount, name, source, s3, mode=None):
    destination = "{}/{}".format(mount, name)
    if mode:
        assert mode in ['File', 'Pipe']
    processing_input = ProcessingInput(
        source=source,
        destination=destination,
        input_name=name,
        # s3_data_type='S3Prefix',
        s3_input_mode=mode or 'File',
        # s3_data_distribution_type='FullyReplicated',
        # s3_compression_type='None'
    )
    path = get_local_path(source)
    if path:
        if not os.path.exists(path):
            raise ValueError(
                "Local path [{}]: [{}] does not exist".format(name, source))
        if os.path.isfile(path):
            basename = os.path.basename(path)
            path_argument = "{}/{}".format(destination, basename)
        elif os.path.isdir(path):
            path_argument = destination
        else:
            raise ValueError(
                "Local path [{}] is neither file nor directory".format(source))
    else:
        if is_s3_file(source, s3=s3):
            basename = os.path.basename(source)
            path_argument = "{}/{}".format(destination, basename)
        else:
            path_argument = destination
    return processing_input, path_argument


def json_converter(o):
    if isinstance(o, datetime.datetime):
        return o.isoformat()  # __str__()
    else:
        raise ValueError("unknown: {}".format(o))


def process(
    session: SagemakerSession,
    role,
    script,
    # script_source,
    inputs=None,
    outputs=None,
    dependencies=None,
    requirements=None,
    configuration_script=None,
    configuration_command=None,
    base_job_name=PROCESSING_JOB_NAME,
    job_name=None,
    image=PROCESSING_IMAGE,
    instance=PROCESSING_INSTANCE,
    volume_size=30,
    runtime_seconds=PROCESSING_RUNTIME_SECONDS,
    output_mount=OUTPUT_MOUNT,
    input_mount=INPUT_MOUNT,
    module_mount=MODULE_MOUNT,
    python='python3',
    wait=True,
    logs=True,
    arguments=None,
    tags=None,
    output_json=None,
    env=None
):
    iam = session.boto_session.client('iam')
    role = ensure_processing_role(iam=iam, role_name=role)
    if inputs is None:
        inputs = {}
    if outputs is None:
        outputs = {}
    if dependencies is None:
        dependencies = {}
    if tags is None:
        tags = {}
    else:
        tags = tags.copy()
    if arguments is None:
        arguments = {}
    else:
        arguments = arguments.copy()
    # if module_mount is not None and len(module_mount)> 0:
    #    command = ["PYTHONPATH={module_mount};${{PYTHONPATH}}".format(module_mount=module_mount), "python3"]
    # else:
    #    command = ['python3']
    command = ['sh']
    path_arguments = {}
    processing_inputs = []
    s3 = session.boto_session.client('s3')
    for name, source in inputs.items():
        processing_input, path_argument = make_processing_input(
            mount=input_mount,
            name=name,
            source=source.local,
            mode=source.mode,
            s3=s3
        )
        processing_inputs.append(processing_input)
        path_arguments[name] = path_argument
    for name, source in dependencies.items():
        processing_input, path_argument = make_processing_input(
            mount=module_mount,
            name=name,
            source=source,
            s3=s3
        )
        processing_inputs.append(processing_input)
        path_arguments[name] = path_argument

    script_remote = "{}/{}".format(module_mount, os.path.basename(script))
    processing_inputs.append(
        ProcessingInput(
            source=script,
            destination=module_mount,
            input_name="aws_sagemaker_remote_script",
            # s3_data_type='S3Prefix',
            s3_input_mode='File',
            # s3_data_distribution_type='FullyReplicated',
            # s3_compression_type='None'
        )
    )
    if env:
        env = env.copy()
    else:
        env = {}
    env.update({
        "AWS_SAGEMAKER_REMOTE_MODULE_MOUNT": module_mount,
        "AWS_SAGEMAKER_REMOTE_PYTHON": python,
        "AWS_SAGEMAKER_REMOTE_SCRIPT": script_remote
    })

    if requirements:
        requirements_remote = "{}/requirements_txt/{}".format(
            module_mount, 'requirements.txt')
        env['AWS_SAGEMAKER_REMOTE_REQUIREMENTS'] = requirements_remote
        processing_inputs.append(
            ProcessingInput(
                source=requirements,
                destination="{}/requirements_txt".format(module_mount),
                input_name="aws_sagemaker_remote_requirements",
                s3_input_mode='File',
            )
        )

    if configuration_script:
        configuration_script_remote = "{}/{}".format(
            module_mount, os.path.basename(configuration_script))
        env['AWS_SAGEMAKER_REMOTE_CONFIGURATION_SCRIPT'] = configuration_script_remote
        processing_inputs.append(
            ProcessingInput(
                source=configuration_script,
                destination=module_mount,
                input_name="aws_sagemaker_remote_configuration_script",
                s3_input_mode='File'
            )
        )

    if configuration_command and len(configuration_command) > 0:
        env['AWS_SAGEMAKER_REMOTE_CONFIGURATION_COMMAND'] = configuration_command

    tags["Source"] = 'aws-sagemaker-remote'
    tags["BaseJobName"] = base_job_name
    tags = make_tags(tags)
    print("Tags: {}".format(tags))
    processor = ScriptProcessor(
        role,
        image_uri=image,
        instance_count=1,
        instance_type=instance,
        command=command,
        volume_size_in_gb=volume_size,
        # volume_kms_key=None,
        # output_kms_key=None,
        max_runtime_in_seconds=runtime_seconds,
        base_job_name=base_job_name,
        sagemaker_session=session,
        env=env,
        tags=tags
        # network_config=None
    )
    processing_outputs = []
    for name, dest in outputs.items():
        # todo: move into PathArgument class
        if not ((not dest.remote) or dest.remote == 'default' or dest.remote.startswith('s3://')):
            raise ValueError("Argument [{}] must be either `default` or an S3 url (`s3://...`). Value given was [{}].".format(
                variable_to_argparse("{}_s3".format(name)), dest.remote))
        source = "{}/{}".format(output_mount, name)
        if dest.mode:
            assert dest.mode in ['EndOfJob', 'Continuous']
        processing_outputs.append(
            ProcessingOutput(
                source=source,
                destination=dest.remote if dest.remote and dest.remote != 'default' else None,
                output_name=name,
                s3_upload_mode=dest.mode or 'EndOfJob'
            ))
        path_arguments[name] = source

    ensure_eol(PROCESSING_SCRIPT)
    code = Path(PROCESSING_SCRIPT).as_uri()
    if job_name is None or len(str(job_name).strip()) == 0:
        job_name = None
    else:
        job_name = str(job_name).strip()

    arguments.update(path_arguments)
    processor.run(
        code=code,
        inputs=processing_inputs,
        outputs=processing_outputs,
        wait=False,
        logs=logs,
        job_name=job_name,
        arguments=sagemaker_arguments(vargs=arguments)
    )
    job = processor.latest_job
    if output_json:
        obj = job.describe()
        print("Describe: {}".format(obj))
        os.makedirs(os.path.dirname(
            os.path.abspath(output_json)), exist_ok=True)
        with open(output_json, 'w') as f:
            json.dump(obj, f, default=json_converter, indent=4)

    if wait:
        job.wait(logs=logs)
    return processor



from botocore.exceptions import ClientError
SAGEMAKER_FULL_ACCESS = 'arn:aws:iam::aws:policy/AmazonSageMakerFullAccess'


def is_boto_exception(e, code):
    return (
        hasattr(e, 'response')
        and 'Error' in e.response
        and 'Code' in e.response['Error']
        and e.response['Error']['Code'] == code
    )


def ensure_role(iam, role_name, description, policies, trust):
    role = get_role_by_name(iam, role_name=role_name)
    if role is None:
        role = create_role(iam=iam, role_name=role_name,
                           description=description, policies=policies, trust=trust)
    return role['Arn']


def get_role_by_name(iam, role_name):
    try:
        response = iam.get_role(
            RoleName=role_name
        )
        return response['Role']
    except ClientError as e:
        if is_boto_exception(e, 'NoSuchEntity'):
            return None
        else:
            raise e


def create_role(iam, role_name, description, policies, trust):
    print("Creating role [{}]".format(role_name))
    response = iam.create_role(
        # Path='string',
        RoleName=role_name,
        AssumeRolePolicyDocument=trust.strip(),
        Description=description,
        # MaxSessionDuration=123,
        # PermissionsBoundary='string',
        Tags=[
            {
                'Key': 'Source',
                'Value': 'aws-ecs-remote'
            },
        ]
    )
    role = response['Role']
    for policy in policies:
        response = iam.attach_role_policy(
            RoleName=role_name,
            PolicyArn=policy
        )
    return role
